from datetime import datetime
from typing import Optional

from fastapi import HTTPException, status
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session
from sqlalchemy.orm import joinedload
from sqlalchemy import desc
from api.db.database import get_db
from api.core.base.services import Service
from api.utils.success_response import success_response
from api.v1.models.coin_token import Token
from api.v1.models.task import TaskStatus, TaskPriority
from api.v1.models.task_association import TaskSubmission, TaskUserAssociation
from api.v1.models.wallet import TokenBalance, Wallet
from api.v1.schemas import task
from api.v1.models import Task
from api.v1.schemas.task import TaskBase

from fastapi import Depends
from fastapi_pagination.ext.sqlalchemy import paginate as pagination
from fastapi_pagination import Params

class TaskService(Service):
    """Task service"""

    def create(self):
        return super().create()

    def delete(self, db, task_id):
        task = db.query(Task).filter(Task.id == task_id).first()

        if not task:
            return {"message": "Task does not exist"}
        else:
            delete = db.delete(task)
            db.commit()
            return {"message": "Task deleted successfully", 
                    "status_code": 200}
            return success_response(
                            status_code=status.HTTP_200_OK,
                            message="Task deleted successfully",
                            data=jsonable_encoder(delete),
                        )

    def fetch(self, db:Session, task_id):
        task = db.query(Task).filter(Task.id == task_id).first()
        if not task:
            return {
                "message": "Task not found.",
                "status_code": 404,
            }
        return TaskBase(**task.__dict__).dict()

    def fetch_all(self, db: Session):
        all_tasks = db.query(Task).all()
        return all_tasks
        
    def fetch_all_partner_task(self, db: Session, user_id):
        all_tasks = db.query(Task).filter(Task.user_id==user_id).all()
        return all_tasks

    def update(self, db, schema, task_id):
        update_task = task_service.update_task(db, schema, task_id)
        return update_task

    def create_task(self, db: Session, schema: task.TaskCreate):

        task_data = schema.model_dump()

        # Convert enum fields
        if "status" in task_data and task_data["status"] is not None:
            task_data["status"] = TaskStatus(task_data["status"])

        if "priority" in task_data and task_data["priority"] is not None:
            task_data["priority"] = TaskPriority(task_data["priority"])

        # Create Task schema object
        # Explicitly handle optional fields
        task_data = {
            k: v
            for k, v in task_data.items()
            if v is not None
            or k in ["description", "assigned_user_id", "required_user_role", "image_url"]
        }

        # Create the Task using the schema
        task = Task(**task_data)

        """Creates a new task"""
        # task = Task(**schema.dict(self))
        db.add(task)
        db.commit()
        db.refresh(task)

        return task

    def update_task(self, db: Session, schema: task.TaskUpdate, task_id=None):
        """Function to update a Task"""

        task = db.query(Task).filter(Task.id == task_id).first()

        if not task:
            raise ValueError("Task not found")

        update_data = schema.dict(exclude_unset=True)

        for key, value in update_data.items():
            setattr(task, key, value)

        db.commit()

        db.refresh(task)

        return task

    def kol_task(self, db):
        kol_tasks = db.query(Task).filter(
            Task.required_user_role == "KOL", Task.status == "IN_PROGRESS"
        )
        return kol_tasks

    def reply_task(self, db):
        reply_tasks_query = db.query(Task).filter(
            Task.required_user_role == "Reply Guy", Task.status != "COMPLETED"
        )
        return reply_tasks_query

    def assign_user_to_task(self, db:Session, task_id, current_user):
        existing_association = (
            db.query(TaskUserAssociation)
            .filter_by(task_id=task_id, user_id=current_user.id)
            .first()
        )

        # checks if user is already assigned
        if existing_association:
            return {
                "message": "User is already assigned to this task",
                "status_code": 400,
            }

        # Create a TaskUserAssociation
        task_user_association = TaskUserAssociation(
            task_id=task_id, user_id=current_user.id
        )

        participant_count = db.query(Task).filter(Task.id == task_id).first()
        if not participant_count:
            return {
                "message": "Task not found"
            }
        participant_count.participants += 1
        

        # Add the association to the database
        db.add(task_user_association)
        db.commit()
        db.refresh(task_user_association)
        db.refresh(participant_count)

        return {"message": "User successfully assigned to the task", "status_code": 200}

    def task_association(self, db, task_id, current_user):
        # Checks if user is assigned to a task
        task_user_association = (
            db.query(TaskUserAssociation)
            .filter_by(task_id=task_id, user_id=current_user.id)
            .first()
        )

        if not task_user_association:
            raise HTTPException(
                status_code=404, detail="User is not assigned to this task"
            )
        else:
            return task_user_association

    def task_submission(self, db:Session, task_user_association, links, task_id):
        task_status = db.query(Task).filter(Task.id == task_id).first()
        status = jsonable_encoder(task_status.status)
        if status == "Completed":
            return{
                "message": "This Task is no longer accepting submissions <br> Task Completed!",
            }
        
        else:
            submissions = []
            for link in links:
                task_submission = TaskSubmission(
                    task_user_id=task_user_association.id, link=link
                )
                db.add(task_submission)
                submissions.append(task_submission)

            db.commit()

            return {
                "message": "Links successfully submitted",
                "submitted_links": [submission.link for submission in submissions],
            }

    def task_query(self, db:Session, current_user):
        my_tasks_query = (
            db.query(Task)
            .join(TaskUserAssociation, Task.id == TaskUserAssociation.task_id)
            .filter(TaskUserAssociation.user_id == current_user)
        )

        return my_tasks_query
    

    def fetch_user_submissions(self, db, task_user_association):
        submissions = (
            db.query(TaskSubmission)
            .filter_by(task_user_id=task_user_association.id)
            .all()
        )
        #print(submissions)
        return submissions
    

    def fetch_users_in_task(self, db, task_id):
        task_associations = db.query(TaskUserAssociation).filter(TaskUserAssociation.task_id == task_id).all()
        #print(task_associations)
        if not task_associations:
            return None
        return task_associations

    
    def fetch_all_task_submissions(self, db, task_id):
        submissions = (
        db.query(TaskSubmission)
        .join(TaskUserAssociation, TaskUserAssociation.id == TaskSubmission.task_user_id)
        .filter(TaskUserAssociation.task_id == task_id)
        .all()
    )

        if not submissions:
            return {
                "message": "No submissions found for this task.",
                "status_code": 200,
            }

        # Format the response
        formatted_submissions = [
            {
                "submission_id": submission.id,
                "user_id": submission.task_user.user_id,
                "task_user_id": submission.task_user_id,
                "link": submission.link,
                "is_verified": submission.is_verified,
                "submitted_at": submission.submitted_at,
                "verified_at": submission.verified_at,
            }
            for submission in submissions
        ]

        return {
            "message": "Submissions fetched successfully.",
            "data": formatted_submissions,
            "status_code": 200,
        }

    
    def verify_submission(self, db, submission_id):
        submission = db.query(TaskSubmission).filter(TaskSubmission.id == submission_id).first()
        if not submission:
            return {
                "message": "Submission not found.",
                "status_code": 404,
            }
        if submission.is_verified == True:
            raise HTTPException(
                status_code=404, detail="Submission already approved")

    # Update the submission to verified
        submission.is_verified = True
        submission.is_paid = False
        submission.verified_at = datetime.utcnow()
        db.commit()
        return submission
    
    def partner_permission_verify_submission(self, db, submission_id, user_id):
        submission_detail = db.query(TaskSubmission).filter(TaskSubmission.id == submission_id).options(joinedload(TaskSubmission.task_user)).first()
        if not submission_detail:
            raise HTTPException(status_code=404, detail="Task association not found")
        task_id = submission_detail.task_user.task_id
        find_task = self.fetch(db, task_id)

        if find_task["user_id"] != user_id:
            raise HTTPException(status_code=404, detail="Opps you dont have permission for this resource")

        return True

    def calculate_reward(self, db:Session, task_associations, task: Task, task_id):
        total_verified_links = 0
        user_verified_links = {}

        #Checks if task is completed
        find_task = db.query(Task).filter(Task.id==task_id).first()
        if find_task.status == "COMPLETED":
            return {
                "message": "This task is marked completed"
            }

        # Step 1: Count total verified links for all users
        for association in task_associations:
            # Skip already rewarded associations
            if association.is_rewarded:
                continue

            verified_links_count = db.query(TaskSubmission).filter(
                TaskSubmission.task_user_id == association.id,
                TaskSubmission.is_verified == True,
                TaskSubmission.is_paid == False  # Only consider unpaid submissions
            ).count()
            #print("Verified Links Count for Association:", verified_links_count)
            user_verified_links[association.user_id] = verified_links_count
            total_verified_links += verified_links_count

        if total_verified_links == 0:
            return {
                "message": "No new verified links found for this task.",
                "status_code": 200,
            }

        # Step 2: Calculate rewards
        for association in task_associations:
            user_id = association.user_id
            verified_links = user_verified_links.get(user_id, 0)

            if verified_links > 0:
                # Calculate the user's total reward
                total_reward = (verified_links / total_verified_links) * task.budget

                # Update the user's wallet balance
                wallet = db.query(Wallet).filter(Wallet.user_id == user_id).first()
                wallet_id = wallet.id

                reward_token = db.query(Token).filter(Token.symbol == task.currency).first()
                token_id = reward_token.id

                token_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id,
                                                             TokenBalance.token_id == token_id).first()
                if token_balance:
                    token_balance.amount += total_reward
                else:
                    # Handle case where wallet doesn't exist
                    token_balance = TokenBalance(TokenBalance.wallet_id==wallet_id, 
                                                 TokenBalance.token_id == token_id,
                                                 TokenBalance.amount == total_reward)
                    db.add(token_balance)

                # Update rewards for verified submissions
                verified_submissions = db.query(TaskSubmission).filter(
                    TaskSubmission.task_user_id == association.id,
                    TaskSubmission.is_verified == True,
                    TaskSubmission.is_paid == False
                ).all()

                for submission in verified_submissions:
                    submission.reward = total_reward / verified_links  # Distribute reward equally
                    submission.is_paid = True  # Mark submission as paid

                # Mark the association as rewarded
                association.is_rewarded = True
            
            
            find_task.status = "COMPLETED"


        # Step 3: Commit changes to the database
        db.commit()

        return {
            "message": "Rewards calculated, stored, and wallets updated successfully.",
            "status_code": 200,
        }

    def task_count(self, db:Session):
        count = db.query(Task).count()
        return count
    
    def completed_tasks(self, db:Session):
        count = db.query(Task).filter(Task.status=="COMPLETED").count()
        return count
    
    def pending_tasks(self, db:Session):
        count = db.query(Task).filter(Task.status=="PENDING").count()
        return count

    def user_task_count(self, db:Session, user_id):
        count = db.query(TaskUserAssociation).filter(TaskUserAssociation.user_id==user_id).count()
        return count
    
    def user_pending_count(self, db:Session, user_id):
        count = db.query(TaskUserAssociation).filter(TaskUserAssociation.user_id==user_id, TaskUserAssociation.is_rewarded==False).count()
        return count
    
    def user_completed_count(self, db:Session, user_id):
        count = db.query(TaskUserAssociation).filter(TaskUserAssociation.user_id==user_id, TaskUserAssociation.is_rewarded==True).count()
        return count
    
    def participants_count(self, db:Session, task_id):
        count = db.query(TaskUserAssociation).filter(TaskUserAssociation.task_id == task_id,
                                                     ).count()
        return count

    def serialize_task(self, task):
            return {
                "id": task.id,
                "title": task.title,
                "image_url": task.image_url,
                "description": task.description,
                "status": task.status,
                "created_at": task.created_at.isoformat(),
                "due_date": task.due_date.isoformat() if task.due_date else None,
                "required_user_role": task.required_user_role or "Unknown",
                "budget": task.budget or 0.0,
                "priority":task.priority,
            }
    
    def task_query_profile(self, db, user_id, params: Params = Depends()):
        
        my_tasks_query = (
            db.query(Task)
            .join(TaskUserAssociation, Task.id == TaskUserAssociation.task_id)
            .filter(TaskUserAssociation.user_id == user_id)
        )

        data = pagination(db, my_tasks_query, params=params)

        # Serialize the tasks
        serialized_tasks = [task.to_dict() for task in data.items]

        return {
            "message": "Tasks retrieved successfully",
            "status_code": 200,
            "data": serialized_tasks,
            "pagination": {
                "current_page": data.page,
                "per_page": data.size,
                "total_pages": data.pages,
                "total": data.total,
            },
        }

    """ Patners services """
    def task_partner_count(self, db:Session, user_id):
        count = db.query(Task).filter(Task.user_id==user_id).count()
        return count
    
    def completed_partner_tasks(self, db:Session, user_id):
        count = db.query(Task).filter(Task.status=="COMPLETED", Task.user_id==user_id).count()
        return count
    
    def pending_partner_tasks(self, db:Session, user_id):
        count = db.query(Task).filter(Task.status=="PENDING", Task.user_id==user_id).count()
        return count
    
    def inprogress_partner_tasks(self, db:Session, user_id):
        count = db.query(Task).filter(Task.status=="IN_PROGRESS", Task.user_id==user_id).count()
        return count
    


    
"""Calls Task services"""
task_service = TaskService()
